#include "dpx.hpp"

DPX::DPX() {}

uint32_t DPX::swapEndian(uint32_t value) {
    return ((value & 0xFF000000) >> 24) |
           ((value & 0x00FF0000) >> 8) |
           ((value & 0x0000FF00) << 8) |
           ((value & 0x000000FF) << 24);
}

uint16_t DPX::swapEndian(uint16_t value) {
    return ((value & 0xFF00) >> 8) |
           ((value & 0x00FF) << 8);
}

bool DPX::readHeader(ifstream& file) {
    // Read magic number
    file.read(reinterpret_cast<char*>(&header.magic), sizeof(header.magic));
    
    // Check for both possible magic numbers (big and little endian)
    const uint32_t MAGIC_BE = 0x53445058;  // "SDPX" in big endian
    const uint32_t MAGIC_LE = 0x58504453;  // "XPDS" in little endian
    
    cout << "{ \"dpx_magic_number\" : \"" << hex << header.magic << "\" }" << endl;
    
    if (header.magic == MAGIC_BE) {
        header.isBigEndian = true;
    } else if (header.magic == MAGIC_LE) {
        header.isBigEndian = false;
    } else {
        uint32_t swapped_magic = swapEndian(header.magic);
        cout << "{ \"dpx_swapped_magic_number\" : \"0x" << hex << swapped_magic << "\" }" << endl;
        
        if (swapped_magic == MAGIC_BE) {
            header.isBigEndian = true;
            header.magic = swapped_magic;
        } else if (swapped_magic == MAGIC_LE) {
            header.isBigEndian = false;
            header.magic = swapped_magic;
        } else {
            throw runtime_error("Not a valid DPX file - Invalid magic number");
        }
    }

    // Read image data offset
    file.seekg(4);
    file.read(reinterpret_cast<char*>(&header.imageOffset), sizeof(header.imageOffset));
    if (header.isBigEndian) {
        header.imageOffset = swapEndian(header.imageOffset);
    }

    // Seek to image information offset (768 bytes from start)
    file.seekg(768);
    file.read(reinterpret_cast<char*>(&header.width), sizeof(header.width));
    file.read(reinterpret_cast<char*>(&header.height), sizeof(header.height));
    
    if (header.isBigEndian) {
        header.width = swapEndian(header.width);
        header.height = swapEndian(header.height);
    }

    cout << "{ \"dimensions\" : { \"width\" : " << header.width << ", \"height\" :" << header.height << " }, ";

    // Read image element information (1556 bytes from start)
    file.seekg(1556);
    file.read(reinterpret_cast<char*>(&header.descriptor), sizeof(header.descriptor));
    file.read(reinterpret_cast<char*>(&header.bitDepth), sizeof(header.bitDepth));
    file.read(reinterpret_cast<char*>(&header.imageElements), sizeof(header.imageElements));

    cout << " \"bit_depth\" : " << static_cast<int>(header.bitDepth) << ", ";
    cout << " \"offset\": " << header.imageOffset << " }" << endl;

    return true;
}

void DPX::readImageData(ifstream& file) {
    file.seekg(header.imageOffset);
    
    image = Mat(header.height, header.width, CV_16UC3);
    
    if (header.bitDepth == 10) {
        size_t packedSize = ((header.width * header.height * 3 * 10 + 31) / 32) * 4;
        vector<uint8_t> buffer(packedSize);
        
        file.read(reinterpret_cast<char*>(buffer.data()), packedSize);
        
        uint16_t* imgPtr = reinterpret_cast<uint16_t*>(image.data);
        size_t pixelCount = 0;
        
        for (size_t i = 0; i < packedSize - 3; i += 4) {
            // Extract four bytes
            uint32_t b0 = buffer[i];
            uint32_t b1 = buffer[i + 1];
            uint32_t b2 = buffer[i + 2];
            uint32_t b3 = buffer[i + 3];
            
            // Combine bytes according to endianness
            uint32_t packed;
            if (header.isBigEndian) {
                packed = (b0 << 24) | (b1 << 16) | (b2 << 8) | b3;
            } else {
                packed = (b3 << 24) | (b2 << 16) | (b1 << 8) | b0;
            }
            
            // Extract three 10-bit values
            uint16_t pix1 = (packed >> 22) & 0x3FF;
            uint16_t pix2 = (packed >> 12) & 0x3FF;
            uint16_t pix3 = (packed >> 2) & 0x3FF;
            
            // Scale to 16-bit
            pix1 <<= 6;
            pix2 <<= 6;
            pix3 <<= 6;
            
            if (pixelCount < header.width * header.height * 3) {
                size_t row = (pixelCount / 3) / header.width;
                size_t col = (pixelCount / 3) % header.width;
                size_t channel = pixelCount % 3;
                
                size_t idx = row * header.width * 3 + col * 3 + (2 - channel);
                imgPtr[idx] = pix1;
                
                if (pixelCount + 1 < header.width * header.height * 3) {
                    channel = (pixelCount + 1) % 3;
                    idx = ((pixelCount + 1) / 3) / header.width * header.width * 3 + 
                          ((pixelCount + 1) / 3) % header.width * 3 + (2 - channel);
                    imgPtr[idx] = pix2;
                }
                
                if (pixelCount + 2 < header.width * header.height * 3) {
                    channel = (pixelCount + 2) % 3;
                    idx = ((pixelCount + 2) / 3) / header.width * header.width * 3 + 
                          ((pixelCount + 2) / 3) % header.width * 3 + (2 - channel);
                    imgPtr[idx] = pix3;
                }
            }
            
            pixelCount += 3;
        }
    } else {
        // Handle other bit depths (8, 12, 16)
        size_t bytesPerPixel = (header.bitDepth + 7) / 8;
        vector<uint8_t> buffer(header.width * header.height * 3 * bytesPerPixel);
        
        file.read(reinterpret_cast<char*>(buffer.data()), buffer.size());
        
        uint16_t* imgPtr = reinterpret_cast<uint16_t*>(image.data);
        
        for (int y = 0; y < header.height; y++) {
            for (int x = 0; x < header.width; x++) {
                for (int c = 0; c < 3; c++) {
                    size_t idx = (y * header.width * 3 + x * 3 + c) * bytesPerPixel;
                    uint16_t pixel = 0;
                    
                    switch (header.bitDepth) {
                        case 8:
                            pixel = buffer[idx] << 8;
                            break;
                        case 12:
                            if (header.isBigEndian) {
                                pixel = ((uint16_t)buffer[idx] << 8) | buffer[idx + 1];
                            } else {
                                pixel = ((uint16_t)buffer[idx + 1] << 8) | buffer[idx];
                            }
                            pixel = pixel << 4;
                            break;
                        case 16:
                            if (header.isBigEndian) {
                                pixel = ((uint16_t)buffer[idx] << 8) | buffer[idx + 1];
                            } else {
                                pixel = ((uint16_t)buffer[idx + 1] << 8) | buffer[idx];
                            }
                            break;
                    }
                    
                    imgPtr[y * header.width * 3 + x * 3 + (2 - c)] = pixel;
                }
            }
        }
    }
}

bool DPX::read(const string& filename) {
    ifstream file(filename, ios::binary);
    if (!file.is_open()) {
        throw runtime_error("Could not open file: " + filename);
    }

    if (!readHeader(file)) {
        return false;
    }

    readImageData(file);
    return true;
}

Mat DPX::getMat() const {
    return image;
}

Mat DPX::getMatBitDepth(int targetBitDepth) const {
    if (targetBitDepth != 8 && targetBitDepth != 10) {
        throw invalid_argument("Target bit depth must be 8 or 10");
    }

    Mat result;
    if (targetBitDepth == 8) {
        image.convertTo(result, CV_8UC3, 1.0/256.0);
    } else if (targetBitDepth == 10) {
        image.convertTo(result, CV_16UC3, 1.0/64.0);
    }
    return result;
}

int DPX::getWidth() const { 
    return header.width; 
}

int DPX::getHeight() const { 
    return header.height; 
}

int DPX::getBitDepth() const { 
    return header.bitDepth; 
}